
-----------------------------------------------------------------
-- Machine Stack include file.
-----------------------------------------------------------------

-- Public Domain.  "Copy Left."

-- Requires Euphoria v4.0

-- Originally designed by James Jonathan Cook (jmsck55@gmail.com)
-- Created: 3/31/2014

-- Last updated by James Jonathan Cook (jmsck55@gmail.com)
-- Updated: 5/22/2014
-- Updated: 12/8/2015
-- Updated: 12/14/2015
-----------------------------------------------------------------

-- STATUS:

-- done. I'll develop this, let someone else maintain it.

-- * needs someone to impliment type checking.

-- REFERENCE:

-- LIFO, last in, first out.

-- resizeable stack, little endian.

-- For reference:
-- http://en.wikipedia.org/wiki/Stack_(abstract_data_type)

-- For information on Euphoria v4.0
-- http://OpenEuphoria.org

-- This is Euphoria code.
-----------------------------------------------------------------

-- VARIABLES AND CONSTANTS:

sequence mystack

atom min_size
atom max_size

-- To do: force min_size and max_size.

min_size = 0 -- has to be at least 0, expecting a large value.
max_size = floor(1073741823 / 4) -- largest integer value: 1073741823
-- max_size is to warn the user that the stack is getting too big.

-- constant first = 1 -- in Euphoria, first is always 1.  In C/C++ it is 0.

atom stackpointer -- has to be greater than or equal to 1
atom stackorigin -- has to be equal or less than length(mystack)
-- when stackorigin is:
--  mystack[stackorigin] = 0
--  stackorigin -= 1
-- data is lost off the bottom of the stack

object error

public procedure condense()
    error = 0
    mystack = mystack[stackpointer..stackorigin]
    stackpointer = 1
    stackorigin = length(mystack)
end procedure

-- CONSTRUCTOR:

public procedure init_seq(sequence s)
-- done.
-- initialize with sequence
    error = 0
    mystack = s
    stackpointer = 1
    stackorigin = length(s)
end procedure

public procedure init_length(atom len)
-- done.
-- initialize with length
    error = 0
    mystack = repeat(0, len)
    stackpointer = len + 1
    stackorigin = len
end procedure

public procedure init()
-- done.
-- initialize with min_size length
    init_length(min_size)
end procedure

init() -- initiates mystack

public function get_stack()
-- done.
-- makes a new sequence
-- return the sequence for the data on the stack
    --error = 0
    condense()
    return mystack
end function

-- PACK AND UNPACK:

public function packstack()
-- done.
    error = 0
    return {stackorigin, stackpointer, mystack}
end function

public procedure unpackstack(sequence packed)
-- done.
    error = 0
    stackorigin = packed[1] -- like a double ended queue.
    stackpointer = packed[2]
    mystack = packed[3]
end procedure

-- NEW AND SET:

public function new_stack()
-- done.
-- make a new stack
    --error = 0
    init()
    return packstack()
end function

public procedure set_stack(sequence s)
-- done.
-- set stack with sequence returned from new_stack() or packstack()
    --error = 0
    unpackstack(s)
end procedure

-- IS_EMPTY:

public function is_empty()
-- done.
    error = 0
    -- returns true or false (1 or 0)
    -- update:
    return stackpointer - 1 = stackorigin
end function

-- SIZE AND RESIZE:

public function stack_size()
-- done.
    error = 0
    -- portable to C/C++ for unsigned ints:
    -- r = iend - (ibegin - 1)
    return stackorigin - (stackpointer - 1)
end function

public procedure stack_resize(atom size)
-- done.
-- stack_resize() reallocates sequence mystack
-- if size is positive, make storage at the top of the stack, trim from the bottom of the stack
-- if size is negative, make storage at the bottom of the stack, trim from the top of the stack
    error = 0
    atom stacksize
    stacksize = stack_size()
    if size > -1 then
	-- makes storage at the top of the stack, truncates at the bottom of the stack
	if size > stacksize then
	    mystack = repeat(0,size - stacksize) & mystack[stackpointer..stackorigin]
	    stackpointer = size - stacksize + 1
	    stackorigin = length(mystack) -- (stackpointer - 1) + stacksize
	else
	    mystack = mystack[stackpointer..stackorigin - (size - stacksize)]
	    stackpointer = 1
	    stackorigin = length(mystack)
	end if
    else
	size = -size
	-- makes storage at the bottom of the stack, truncates at the top of the stack
	if size > stacksize then
	    mystack = mystack[stackpointer..stackorigin] & repeat(0,size - stacksize)
	    stackpointer = 1
	    stackorigin = stacksize
	else
	    mystack = mystack[stackpointer + (size - stacksize)..stackorigin]
	    stackpointer = 1
	    stackorigin = length(mystack)
	end if
    end if
end procedure

-- STORAGE functions:

public function storage_size()
-- done.
    atom front, back
    error = 0
    front = stackpointer - 1
    back = length(mystack) - stackorigin
    return {front, back}
end function

public procedure add_storage(atom front, atom back)
-- done.
    error = 0
    mystack = repeat(0,front) & mystack & repeat(0,back)
    stackpointer += front
    stackorigin += front
end procedure

public procedure subtract_storage(atom front, atom back)
-- done.
    error = 0
    mystack = mystack[front + 1..$ - back]
    stackpointer -= front
    stackorigin -= front
end procedure

-- SHIFT

public procedure shift(atom front, atom back)
-- done. ??? test.
-- virtual shift, in all intents and purposes, it is a shift, except in mystack
-- data type, it is more like a fast_shift.

-- done. it can delete a lot of data, so use carefully
-- equivalent to a shift, but faster to delete the data at the bottom of the stack
    error = 0
    
    -- fast Euphoria code:
    mystack[stackpointer..stackpointer + (front - 1)] = 0
    mystack[stackorigin - (back - 1)..stackorigin] = 0
    
    stackpointer += front
    stackorigin -= back
    
    -- another way of doing it, for C/C++
--  for i = 1 to howmany do
--      mystack[stackorigin] = 0
--      stackorigin -= 1
--  end for
    
    
end procedure


-- TRUNCATE:

public procedure truncate_by(atom howmany) -- truncate by howmany
-- done.
    -- similar to shift(), but reallocates stack.
    error = 0
    -- take howmany off of the bottom of the stack
    mystack = mystack[stackpointer..stackorigin - howmany]
    -- update:
    stackpointer = 1
    stackorigin = length(mystack)
end procedure

-- ITERATORS:

public function stack_begin()
-- done.
    error = 0
    return stackpointer
end function

public function stack_end()
-- done.
    error = 0
    return stackorigin
end function

-- Functions that we need:

public function top()
-- done.
    error = 0
    return mystack[stackpointer]
end function

public function bottom() -- bottom of the stack, or last element on the stack
-- done.
    error = 0
    return mystack[stackorigin]
end function

public function at(atom index) -- index from 1 to stack_size()
-- done.
    error = 0
    return mystack[(stackpointer - 1) + index]
end function

-- PUSH and POP functions:

public procedure push_range(sequence vals) -- pushing a sequence of values onto the stack
--done.
    error = 0
    if stackpointer > length(vals) then
	-- offset stackpointer going in the negative direction
	mystack[stackpointer - length(vals)..stackpointer - 1] = vals
	stackpointer -= length(vals)
	
    else -- growing the stack.
	-- put some values in mystack until it fills up
	mystack[1..stackpointer - 1] = vals[$ - (stackpointer - 1) + 1..$]
	if stackorigin != length(mystack) then
	    mystack = mystack[1..stackorigin] -- frees memory
	    --stackpointer = 1
	    stackorigin = length(mystack)
	end if
	
	-- This is the correct Euphoria code: (dont delete, others might find it useful)
	-- might need a +1 ???
	mystack = vals[1..$ - (stackpointer - 1)] & mystack
	
	-- in a for loop, it would look like this:
	--for i = length(vals) - (stackpointer - 1) to 1 by -1 do
	    --mystack = prepend(mystack, vals[i])
	--end for
	
	stackorigin += (length(vals) - (stackpointer - 1))
	stackpointer = 1
	
	if max_size < length(mystack) then
	    -- mystack is larger than max_size
	    -- try, set_size(max_size) OR change max_size.
	    error = 3
	end if
    end if
end procedure

--updated:
public procedure push(object val) -- pushing one value onto the stack
-- done.
    error = 0
    if stackpointer > 1 then
	stackpointer -= 1
	mystack[stackpointer] = val
	
    elsif stackpointer = 1 then -- growing the stack.
	if stackorigin != length(mystack) then
	    mystack = mystack[stackpointer..stackorigin] -- frees memory
	    --stackpointer = 1
	    stackorigin = length(mystack)
	end if
	
	mystack = prepend(mystack, val)
	--stackpointer = 1
	stackorigin += 1
	
	if max_size < length(mystack) then
	    -- mystack is larger than max_size
	    -- try, set_size(max_size) OR change max_size.
	    error = 3
	end if
    else
	error = 1 -- error, cannot push item because of wrong value for stackpointer
    end if
end procedure

public function pop_range(atom howmany)
--done.
    sequence r
    error = 0
    
    if (stackpointer - 1) + howmany <= stackorigin then -- Euphoria code
	r = mystack[stackpointer..(stackpointer - 1) + howmany]
	mystack[stackpointer..(stackpointer - 1) + howmany] = 0 -- free memory
	stackpointer += howmany
	
    elsif (stackpointer - 1) + (howmany - 1) = stackorigin then
	r = mystack[stackpointer..stackorigin] -- return all items on stack
	init() -- stack is empty, re-initialize
	error = (-2) -- had to make a new stack, all items on stack are returned
    else
	error = 2
	r = {}
    end if
    return r
end function

public function pop() -- LIFO. Last in, First out.
-- done.
    object r
    error = 0
    if stackpointer <= stackorigin then -- Euphoria code
	r = mystack[stackpointer]
	mystack[stackpointer] = 0 -- free memory
	stackpointer += 1 -- advance stack pointer toward bottom of stack
	
    elsif stackpointer - 1 = stackorigin then -- free up memory
	-- shrinking the stack.
	r = mystack[stackorigin] -- return the last element in stack
	init() -- stack is empty, re-initialize
	error = (-1) -- had to make new stack, let user know (-1)
    else
	-- ???, should never happen
	error = 2 -- no items to pop
	r = {}
    end if
    
    return r
end function

public procedure push_bottom_range(sequence vals)
--done.
    error = 0
    --here--
    if stackorigin < length(mystack) - length(vals) + 1 then
	-- offset stackorigin going in the positive direction
	mystack[stackorigin + 1..stackorigin + length(vals)] = vals
	stackorigin += length(vals)
	
    else -- growing the stack.
	-- put some values in mystack until it fills up
	mystack[stackorigin + 1..$] = vals[1..length(mystack) - stackorigin]
	if stackpointer != 1 then
	    mystack = mystack[stackpointer..$] -- frees memory
	    stackpointer = 1
	    stackorigin = length(mystack)
	end if
	
	mystack = mystack & vals[length(mystack) - stackorigin + 1..$]
	
	-- in a for loop, it would look like this:
	--for i = length(mystack) - stackorigin + 1 to length(vals) do
	    --mystack = append(mystack, vals[i])
	--end for
	
	stackorigin = length(mystack)
	
	if max_size < length(mystack) then
	    -- mystack is larger than max_size
	    -- try, set_size(max_size) OR change max_size.
	    error = 3
	end if
    end if
end procedure

--updated:
public procedure push_bottom(object val)
--done.
    error = 0
    if stackorigin < length(mystack) then
	stackorigin += 1
	mystack[stackorigin] = val
	
    elsif stackorigin = length(mystack) then -- growing the stack.
	if stackpointer != 1 then
	    mystack = mystack[stackpointer..stackorigin] -- frees memory
	    stackpointer = 1
	    stackorigin = length(mystack)
	end if
	
	mystack = append(mystack, val)
	stackorigin += 1
	
	if max_size < length(mystack) then
	    -- mystack is larger than max_size
	    -- try, set_size(max_size) OR change max_size.
	    error = 3
	end if
    else
	error = 1 -- error, stackorigin has a wrong value
    end if
end procedure

public function pop_bottom_range(atom howmany)
--done.
    sequence r
    error = 0
    
    if stackpointer <= (stackorigin + 1) - howmany then -- Euphoria code
	r = mystack[(stackorigin + 1) - howmany..stackorigin]
	mystack[(stackorigin + 1) - howmany..stackorigin] = 0 -- free memory
	stackorigin -= howmany
	
    elsif stackpointer - 1 = stackorigin - (howmany - 1) then
	r = mystack[stackpointer..stackorigin] -- return all items on stack
	init() -- stack is empty, re-initialize
	error = (-2) -- had to make a new stack, all items on stack are returned
    else
	error = 2
	r = {}
    end if
    return r
end function

public function pop_bottom()
--done.
    object r
    error = 0
    if stackpointer <= stackorigin then -- Euphoria code
	r = mystack[stackorigin]
	mystack[stackorigin] = 0 -- free memory
	stackorigin -= 1 -- subtract stackorigin toward top of stack
	
    elsif stackpointer - 1 = stackorigin then -- free up memory
	-- shrinking the stack.
	r = mystack[stackpointer] -- return the first element in stack
	init() -- stack is empty, re-initialize
	error = (-1)
    else
	-- ???, should never happen
	error = 2 -- no items to pop
	r = {}
    end if
    
    return r
end function

-- INSERT and ERASE

--updated:
public procedure stack_insert(object items, atom where)
-- done.
    atom count
    error = 0
    -- 'where' should be between stackpointer and stackorigin.
    mystack = splice(mystack, items, where) -- uses new 4.0 'splice' function
    if atom(items) then
	count = 1
    else
	count = length(items)
    end if
    -- adjust stackpointer and stackorigin
    if where <= stackpointer then
	stackpointer += count
    end if
    if where <= stackorigin then
	stackorigin += count
    end if
end procedure

public procedure stack_erase(atom ibegin, atom iend)
-- done.
-- ibegin and iend must be between stackpointer and stackorigin.
    atom count
    error = 0
    mystack = remove(mystack, ibegin, iend) -- uses new 4.0 'remove' function
    count = iend - (ibegin - 1)
    
    -- adjust stackpointer and stackorigin
    
    --most cases:
    stackorigin -= count
    
end procedure


-- EXTRA:

-- These two functions are very similar, but have much different results.

-- Good for popping through "whitespace" such as " \n\r\t"

public function pop_until(sequence s)
-- done.
-- pop_until element in s is reached
    sequence result
    error = 0
    result = {}
    while stackpointer <= stackorigin do --length(mystack) do
	if find(top(), s) then -- notice 'find'
	    exit
	end if
	result = append( result, pop() )
    end while
    
    return result
end function

public function pop_while(sequence s)
-- done.
-- pop_while elements are of s
    sequence result
    error = 0
    result = {}
    while stackpointer <= stackorigin do --length(mystack) do
	if not find(top(), s) then -- notice 'not find'
	    exit
	end if
	result = append( result, pop() )
    end while
    
    return result
end function


